//! ```
//! use core::str::FromStr;
//! use std::net::{IpAddr, Ipv4Addr, SocketAddr};
//!
//! use socket_address::Address;
//!
//! let addr = Address::from_str("0.0.0.0:80").unwrap();
//! assert_eq!(addr, Address::Network(SocketAddr::new(IpAddr::V4(Ipv4Addr::new(0, 0, 0, 0)), 80)));
//! ```
//!
//! ```
//! use core::str::FromStr;
//! use std::path::PathBuf;
//!
//! use socket_address::Address;
//!
//! let addr = Address::from_str("unix:myapp.sock").unwrap();
//! assert_eq!(addr, Address::UnixSocket("myapp.sock".into()));
//! ```
use core::fmt;
use core::str::FromStr;
use std::net::SocketAddr;
use std::path::PathBuf;

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum Address {
	Network(SocketAddr),
	UnixSocket(PathBuf)
}

impl FromStr for Address {
	type Err = InvalidListenAddress;

	fn from_str(s: &str) -> Result<Self, Self::Err> {
		match s.strip_prefix("unix:") {
			Some(v) => Ok(Self::UnixSocket(v.into())),
			None => match SocketAddr::from_str(s) {
				Ok(v) => Ok(Self::Network(v)),
				Err(_) => Err(InvalidListenAddress(s.to_owned()))
			}
		}
	}
}

impl fmt::Display for Address {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
		match self {
			Self::Network(addr) => addr.fmt(f),
			Self::UnixSocket(path) => {
				f.write_str("unix:")?;
				path.display().fmt(f)
			}
		}
	}
}

#[derive(Debug, thiserror::Error)]
#[error("'{0}' is neither a valid IP:port address nor a path prefixed with 'unix:'")]
pub struct InvalidListenAddress(String);
